#include "bms_converter.h"
#include <string.h>

#define DIV_BY_F( a, b ) (((double)(a)) / ((double)(b)))
#define MUL_BY_F( a, b ) (((double)(a)) * ((double)(b)))

BmsConverter::BmsConverter() {
  this->bms_path = NULL;
  this->bms = NULL_BMS;
  this->master_volume = 100;
  this->callback = NULL;
  this->wav_file_empty_callback = NULL;
  this->wnd = NULL;
  this->wnd_callback = NULL;
  this->wnd_wav_empty = NULL;
  this->wnd_wav_file_empty_callback = NULL;
  this->debug = false;
}

BmsConverter::~BmsConverter() {
  delete this->bms;
  delete this->bms_path;
}

void BmsConverter::set_bms_path( char *bms_path ) {
  this->bms_path = new char[strlen( bms_path ) + 1];
  strcpy( this->bms_path, bms_path );
}

void BmsConverter::set_debug( bool flag ) {
  this->debug = flag;
}

BmsBms *BmsConverter::open_bms() {
  if ( this->debug ) { printf( "file open %s\n", this->bms_path ); }
  bool status = BmsLexer::read_file( this->bms_path, &this->bms );
  if ( !status ) { this->bms = NULL_BMS; }
  return this->bms;
}

void BmsConverter::set_master_volume( int volume ) {
  this->master_volume = volume;
}

void BmsConverter::set_callback( bool (* callback)( int ) ) {
  this->callback = callback;
}

void BmsConverter::set_wnd_callback( TtWindow *wnd, bool (TtWindow::*wnd_callback)( int ) ) {
  this->wnd = wnd;
  this->wnd_callback = wnd_callback;
}

void BmsConverter::set_wav_file_empty_callback( bool (* wav_file_empty_callback)( char * ) ) {
  this->wav_file_empty_callback = wav_file_empty_callback;
}

void BmsConverter::set_wnd_wav_file_empty_callback( TtWindow *wnd, bool (TtWindow::*wnd_wav_file_empty_callback)( char * ) ) {
  this->wnd_wav_empty = wnd;
  this->wnd_wav_file_empty_callback = wnd_wav_file_empty_callback;
}

TtWav *BmsConverter::convert() {
  bool status;
  int callback_prev = -1;

  if ( this->callback != NULL ) { (* callback)( -1 ); }
  if ( this->wnd_callback != NULL ) { (this->wnd->*this->wnd_callback)( -1 ); }
  
  char bms_dir[MAX_PATH];
  {
    char drive[MAX_PATH], dir[MAX_PATH], fname[MAX_PATH], ext[MAX_PATH];
    _splitpath( this->bms_path, drive, dir, fname, ext );
    sprintf( bms_dir, "%s%s", drive, dir );
    if ( this->debug ) { printf( "%s\n", bms_dir ); }
  }

  int wav_count = this->bms->get_wavs()->get_exist_count();
  int wav_readed = 0;
  int max_data_size_of_wavs = 0;
  TtWav **wavs = new TtWav*[BmsWord::MAX];
  memset( wavs, 0, sizeof(TtWav*) * BmsWord::MAX );
  for ( int i = 0; i < BmsWord::MAX; ++i ) {
    const char *_tmp = this->bms->get_wavs()->at( i );

    if ( _tmp[0] == '\0' ) { wavs[i] = 0; }
    else {
      wavs[i] = new TtWav();
      wavs[i]->set_debug( this->debug );

      char buf[MAX_PATH];
      sprintf( buf, "%s%s", bms_dir, _tmp );
      if ( this->debug ) { printf( "----wav open %s-------------------------\n", buf ); }
      status = wavs[i]->read_file( buf );
      if ( !status ) {
        strcpy( this->last_error_filename, buf );
        if ( this->debug ) { printf( "WAV READ error => %s \n", buf ); }

        // BMSŎgĂ邩ǂ̃`FbNB
        bool used_flag = false;
        bool exchange_flag = false;
        EACH_CHANNEL( this->bms->get_first_channel(), temp_channel ) {
          if ( BmsUtil::is_wav_channel( temp_channel->get_channel_number() ) ) {
            if ( temp_channel->contains( BmsWord( i ) ) ) { used_flag = true; break; }
          }
        }
        if ( used_flag == false ) {
          if ( this->wav_file_empty_callback != NULL ) {
            exchange_flag = (* this->wav_file_empty_callback)( buf );
          }
          else if ( this->wnd_wav_file_empty_callback != NULL ) {
            exchange_flag = (this->wnd_wav_empty->*this->wnd_wav_file_empty_callback)( buf );
          }
          else {
            exchange_flag = false;
          }
        }
        if ( exchange_flag != true ) {
          if ( wavs != NULL ) {
            for ( int i = 0; i < BmsWord::MAX; ++i ) {
              if ( wavs[i] != NULL ) { delete wavs[i]; }
            }
            delete wavs;
          }
          return 0;
        }
      }
      wavs[i]->change_volume( this->master_volume );
      max_data_size_of_wavs = wavs[i]->get_data_size() > max_data_size_of_wavs ? wavs[i]->get_data_size() : max_data_size_of_wavs;
      ++wav_readed;
    }
    if ( this->callback != NULL || this->wnd_callback != NULL ) {
      int now = wav_readed * 100 / wav_count;
      if ( now != callback_prev ) {
        for ( int i = now - callback_prev - 1; i >= 0; --i ) {
          if ( this->callback != NULL ) { (* callback)( now - i ); }
          if ( this->wnd_callback != NULL ) { (this->wnd->*this->wnd_callback)( now - i ); }
        }
        callback_prev = now;
      }
    }
  }
  if ( this->debug ) { printf( "max_data_size_of_wavs is %d\n", max_data_size_of_wavs ); }

  TtWav *all = new TtWav();
  all->set_debug( this->debug );
  all->make_blank();
  const char *bpm_str = this->bms->get_headers()->fetch_str( "BPM" );
  int now_bpm = bpm_str[0] == '\0' ? 120 : atoi( bpm_str );

  if ( this->debug ) { printf( "max position is %d tic\n", (int)MUL_BY_F( this->bms->get_max_position(), DIV_BY_F( 44100 * 60, BmsBar::DEFAULT_LENGTH / 4 * now_bpm ) ) ); }

  all->memory_extend( 
    (int)MUL_BY_F( this->bms->get_max_position(), DIV_BY_F( 44100 * 60, BmsBar::DEFAULT_LENGTH / 4 * now_bpm ) ) * 4 +
    max_data_size_of_wavs +
    44100 * 2 * 2 );

  if ( this->debug ) {
    printf( "max position's data is %d\n",
            (int)MUL_BY_F( this->bms->get_max_position(), DIV_BY_F( 44100 * 60, BmsBar::DEFAULT_LENGTH / 4 * now_bpm ) ) * 4 +
            max_data_size_of_wavs );
  }
  // Om[gȂPc폜鏈
  EACH_KIND_CHANNEL( this->bms->get_first_channel(), temp_channel ) {
    int ch_num = temp_channel->get_channel_number().get_int();
    if ( 181 <= ch_num && ch_num <= 215 || 217 <= ch_num && ch_num <= 251 ) {
      bool on_obj = false;
      BmsWord on_obj_val( "00" );
      for ( int i = 0; i < temp_channel->get_length(); ++i ) {
        if ( on_obj ) {
          if ( temp_channel->at( i ) != "00" ) {
            if ( temp_channel->at( i ) != on_obj_val ) { break; }
            temp_channel->overwrite( i, BmsWord( "00" ) );
            on_obj = false;
          }
        }
        else {
          if ( temp_channel->at( i ) != "00" ) {
            on_obj_val = temp_channel->at( i ); on_obj = true;
          }
        }
      }
      if ( on_obj ) { 
        if ( this->debug ) { printf( "LONG NOTE OBJECT FORMAT error \n" ); }
        if ( all != NULL ) { delete all; }
        if ( wavs != NULL ) {
          for ( int i = 0; i < BmsWord::MAX; ++i ) {
            if ( wavs[i] != NULL ) { delete wavs[i]; }
          }
          delete wavs;
        }
        return NULL;
      }
    }
    // sIuWF͏
    else if ( 109 <= ch_num && ch_num <= 143 || 145 <= ch_num && ch_num <= 180 ) {
      temp_channel->clear();
    }
  }
  
  now_bpm;
  double all_pos = 0.0;
  int max_pos = this->bms->get_max_position();
  for ( int i = 0; i <= max_pos; ++i ) {
    EACH_CHANNEL( this->bms->get_first_channel(), temp_channel ) {
      if ( BmsUtil::is_wav_channel( temp_channel->get_channel_number() ) ) {
        if ( temp_channel->at( i ) == "00" ) { continue; }
        if ( wavs[temp_channel->at( i ).get_int()] == NULL ) { continue; }
        all->mixin_at( (int)all_pos, wavs[temp_channel->at( i ).get_int()] );
      }
    }

    // BPMύX
    if ( this->bms->get_channel( BmsWord( "03" ) ) != NULL_CHANNEL &&
         this->bms->get_channel( BmsWord( "03" ) )->at( i ) != "00" ) {
      if ( this->debug ) { printf("bpm changed %d =>", now_bpm ); }
      char buf[3];
      char *p;
      this->bms->get_channel( BmsWord( "03" ) )->at( i ).get_str_null( buf );
      now_bpm = strtol( buf, &p, 16 );
      if ( buf == p ) {
        if ( this->debug ) { printf( "BPM error => %s\n", buf ); }
        if ( all != NULL ) { delete all; }
        if ( wavs != NULL ) {
          for ( int i = 0; i < BmsWord::MAX; ++i ) {
            if ( wavs[i] != NULL ) { delete wavs[i]; }
          }
          delete wavs;
        }
        return NULL;
      }
      if ( this->debug ) { printf(" %d\n", now_bpm ); }
    }
    
    // gBPMύX
    if ( this->bms->get_channel( BmsWord( "08" ) ) != NULL_CHANNEL &&
         this->bms->get_channel( BmsWord( "08" ) )->at( i ) != "00" ) {
      if ( this->debug ) { printf("exbpm changed %d =>", now_bpm ); }
      char buf[1024];
      sprintf( buf, "BPM%c%c",
               this->bms->get_channel( BmsWord( "08" ) )->at( i ).get_msb(),
               this->bms->get_channel( BmsWord( "08" ) )->at( i ).get_lsb() );
      const char *bpm_str = this->bms->get_headers()->fetch_str( buf );
      char *p;
      now_bpm = strtol( bpm_str, &p, 10 );
      if ( bpm_str == p ) {
        if ( this->debug ) { printf( "ex BPM error => %s %s\n", buf, bpm_str ); }
        if ( all != NULL ) { delete all; }
        if ( wavs != NULL ) {
          for ( int i = 0; i < BmsWord::MAX; ++i ) {
            if ( wavs[i] != NULL ) { delete wavs[i]; }
          }
          delete wavs;
        }
        return NULL;
      }
      if ( this->debug ) { printf(" %d\n", now_bpm ); }
    }
    
    // STOPV[PX
    if ( this->bms->get_channel( BmsWord( "09" ) ) != NULL_CHANNEL &&
         this->bms->get_channel( BmsWord( "09" ) )->at( i ) != "00" ) {
      char buf[1024];
      sprintf( buf, "STOP%c%c",
               this->bms->get_channel( BmsWord( "09" ) )->at( i ).get_msb(),
               this->bms->get_channel( BmsWord( "09" ) )->at( i ).get_lsb() );
      const char *stop_str = this->bms->get_headers()->fetch_str( buf );
      char *p;
      int stop_192 = strtol( stop_str, &p, 10 );
      if ( stop_str == p ) {
        if ( this->debug ) { printf( "ex STOP error => %s %s\n", buf, bpm_str ); }
        if ( all != NULL ) { delete all; }
        if ( wavs != NULL ) {
          for ( int i = 0; i < BmsWord::MAX; ++i ) {
            if ( wavs[i] != NULL ) { delete wavs[i]; }
          }
          delete wavs;
        }
        return NULL;
      }
      all_pos += ((double)(stop_192 * ( BmsBar::DEFAULT_LENGTH / 192 ))) * DIV_BY_F( 44100 * 60, BmsBar::DEFAULT_LENGTH / 4 * now_bpm );
      if ( this->debug ) { printf("STOP %d\n", ((double)(stop_192 * ( BmsBar::DEFAULT_LENGTH / 192 ))) * DIV_BY_F( 44100 * 60, BmsBar::DEFAULT_LENGTH / 4 * now_bpm ) ); }
    }
    
    all_pos += DIV_BY_F( 44100 * 60, BmsBar::DEFAULT_LENGTH / 4 * now_bpm );
    
    if ( this->callback != NULL || this->wnd_callback != NULL ) {
      if ( i == max_pos ) {
          if ( this->callback != NULL ) { (* callback)( 200 ); }
          if ( this->wnd_callback != NULL ) { (this->wnd->*this->wnd_callback)( 200 ); }
      }
      else {
        int now = (int)DIV_BY_F( i * 100, max_pos ) + 100;
        if ( now != callback_prev ) {
          if ( this->callback != NULL ) { (* callback)( now ); }
          if ( this->wnd_callback != NULL ) { (this->wnd->*this->wnd_callback)( now ); }
          callback_prev = now;
        }
      }
    }
  }

  for ( int i = 0; i < BmsWord::MAX; ++i ) {
    if ( wavs[i] != NULL ) { delete wavs[i]; }
  }
  delete wavs;
  
  return all;
}


char *BmsConverter::get_last_error_filename() {
  return this->last_error_filename;
}
